'
'  pb_ftp.bas
'
'  32-bit PowerBasic ftp routines
'  Copyright 1999
'  Don Dickinson
'  ddickinson@basicguru.com
'
'
'  The following are needed by, but not $included in this module:
'     win32api.inc            Power Basic's Win API declaraions
'
'  The following are need by and $included within
'     wsconst.inc             WinSock Constants
'     ftp_cons.inc            FTP Constants and error message routines
'     ftp_help.bas            FTP helper routines
'
'  =======================================
'  Routines available in this module
'  =======================================
'     Function ftpChangeDir      (ByVal hSocket As Long, sDir As String) As Long
'     Function ftpConnect        (sServer As String, sUser As String, sPW As String) As Long
'     Function ftpDeleteFile     (ByVal hSocket As Long, remoteFile As String) As Long
'     Function ftpCurDir         (ByVal hSocket As Long) As String
'     Function ftpGetDirList     (ByVal hSocket As Long, sServer As String, sDirs() As String) As Long
'     Function ftpGetFile        (ByVal hSocket As Long, ByVal hStatus As Long, sServer As String, _
'                                 remoteFile As String, localFile As String) As Long
'     Function ftpGetFileList    (ByVal hSocket As Long, sServer As String, sFiles() As String) As Long
'     Function ftpGetList        (ByVal hSocket As Long, sServer As String, sList() As String) As Long
'     Function ftpMakeDir        (ByVal hSocket as Long, sDir as String) as Long
'     Function ftpPutFile        (ByVal hSocket As Long, ByVal hStatus As Long, _
'                                 remoteFile As String, localFile As String) As Long
'     Sub      ftpQuit           (ByVal hSocket As Long)
'     Function ftpMoveOrRename   (ByVal hSocket as Long, oldFile as String, newFile as String) as Long
'     Sub      ftpSetAscii       (ByVal hSocket As Long)
'     Sub      ftpSetBinary      (ByVal hSocket As Long)
'
'     Function ftpGetLastStatusCode    as Long
'     Function ftpGetLastStatusMsg     as String
'     Function ftpTranslateErrorCode   (ByVal errorCode As Long) As String
'
'  =======================================
'  How the library works - Important Notes
'  =======================================
'
'     >  Call ftpConnect first - this connects to the server. It returns
'        a valid socket handle if successful - Pass this handle to all the
'        other calls. If it fails it returns %INVALID_SOCKET
'
'     >  Call any other commands you wish to.
'
'     >  Most functions return %True if successful or %False if not.
'        If you need more detail about why the call failed, you can
'        use ftpGetLastStatusCode to return the last status code
'        returned from the ftp server (or a dos error code if one occurred)
'        The text error message can be retrieved by calling ftpGetLastStatusMsg
'
'     >  When you're done, call ftpQuit to clean up.
'
'     >  You should never need to call ftpSetBinary or ftpSetAscii
'        as all file operations are done in binary mode. The module calls
'        ftpSetBinary as needed. These will likely not be exported from
'        the final DLL version of this library.
'
'==============================================================================
'
'  Required Include Files
'==============================================================================
$Include "win32api.inc"
$Include "wsconst.inc"
$Include "ftp_cons.inc"
$Include "ftp_help.bas"
'
'  ftpGetLastStatusCode
'  Returns the FTP status code for the last command sent
'  These correspond to the %FTP_ constants in ftp_cons.inc
'==============================================================================
Function ftpGetLastStatusCode Export As Long
   Function = gLastErr
End Function
'
'  ftpGetLastStatusMsg
'  Returns the text message returned by the ftp server
'  during the last communication.
'==============================================================================
Function ftpGetLastStatusMsg Export As String
   Function = gLastMsg
End Function
'
'  ftpDeleteFile
'  Deletes the file specified from the current working directory of the server
'==============================================================================
Function ftpDeleteFile Alias "ftpDeleteFile" _
                  (ByVal hSocket As Long, remoteFile As String) Export As Long
   Dim buffer As String
   Tcp Print hSocket, "DELE " + remoteFile
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
      Function = %True
   Else
      Function = %False
   End If
End Function
'
'  ftpMakeDir
'  Creates a directory.
'==============================================================================
Function ftpMakeDir(ByVal hSocket As Long, sDir As String) Export As Long
   Dim buffer As String
   Tcp Print hSocket, "MKD " + sDir
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = %FTP_DIRECTORY_OK Then
      Function = %True
   Else
      Function = %False
   End If
End Function
'
'  ftpRemoveDir
'  Deletes a directory
'==============================================================================
Function ftpRemoveDir(ByVal hSocket As Long, sDir As String) Export As Long
   Dim buffer As String
   Tcp Print hSocket, "RMD " + sDir
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
      Function = %True
   Else
      Function = %False
   End If
End Function
'
'  ftpMoveOrRename
'  Renames or moves a file
'==============================================================================
Function ftpMoveOrRename(ByVal hSocket As Long, oldFile As String, newFile As String) Export As Long
   Dim buffer As String
   '- Rename From
   Tcp Print hSocket, "RNFR " + oldFile
   Tcp Line hSocket, buffer
   If buffer_status(buffer) = %FTP_ACTION_PENDING Then
      '- Rename To
      Tcp Print hSocket, "RNTO " + newFile
      Tcp Line hSocket, buffer
      If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
         Function = %True
      Else
         Function = %False
      End If
   Else
      Function = %False
   End If
   set_last_status buffer
End Function
'
'  ftpSetAscii
'  turns on ascii transfer mode
'==============================================================================
Sub ftpSetAscii Alias "ftpSetAscii" (ByVal hSocket As Long) Export
   Dim buffer As String
   Tcp Print hSocket, "TYPE A"
   Tcp Line hSocket, buffer
   set_last_status buffer
End Sub
'
'  ftpSetBinary
'  turns on binary transfer mode
'==============================================================================
Sub ftpSetBinary Alias "ftpSetBinary" (ByVal hSocket As Long) Export
   Dim buffer As String
   Tcp Print hSocket, "TYPE I"
   Tcp Line hSocket, buffer
   set_last_status buffer
End Sub
'
'  ftpCurDir
'  returns the current working directory on the server
'  returns "" if an error occurs
'==============================================================================
Function ftpCurDir Alias "ftpCurDir" (ByVal hSocket As Long) Export As String
   Dim buffer As String
   Tcp Print hSocket, "PWD"
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = 257 Then
      Function = Parse$(buffer, Chr$(34), 2)
   Else
      Function = ""
   End If
End Function
'
'  ftpChangeDir
'  Returns %True if it can change the directory
'  Returns %False if not.
'==============================================================================
Function ftpChangeDir Alias "ftpChangeDir" (ByVal hSocket As Long, sDir As String) Export As Long
   Dim buffer As String
   Tcp Print hSocket, "CWD " + Trim$(sDir)
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = 250 Then
      Function = %True
   Else
      Function = %False
   End If
End Function
'
'  ftpGetFile
'  Opens a binary mode connection and downloads the requested file.
'  Returns %True if successful and %False if not.
'  If hStatus is non-zero it is assumed to be a windows handle. As the
'  status of the download changes the caption of this window is changed to
'  be the percent complete of the transfer. If the file already exists then
'  it will be erased and replaced by the downloaded file.
'  Note that this function ALWAYS operates in passive mode - that is, it will
'  always connect to the server's data socket, it doesn't create it's own.
'==============================================================================
Function ftpGetFile Alias "ftpGetFile" (ByVal hSocket As Long, _
                     ByVal hStatus As Long, sServer As String, _
                     remoteFile As String, localFile As String) Export As Long
   Dim hData As Long
   Dim iFF As Long
   Dim iProgress As Long
   Dim iTotal As Long
   Dim zText As Asciiz * 200
   Dim buffer As String
   On Error Resume Next
   iFF = FreeFile
   Kill localFile
   If Dir$(localFile) <> "" Then
      buffer = Format$(Err, "000") + " Local File Can't be deleted"
      Function = %False
   Else
      On Error Resume Next
      Open localFile For Binary Lock Read Write As #iFF
      If Err Then
         buffer = Format$(Err, "000") + " Local File Open Error"
         Function = %False
      Else
         '- Open a passive-mode connection to the server.
         hData = get_passive_handle(hSocket, sServer)
         If hData = %INVALID_SOCKET Then
            buffer = Format$(gLastErr, "000") + " " + gLastMsg
            Function = %False
         Else
            '- Make sure that we are doing a binary transfer
            ftpSetBinary hSocket
            '- Request the file
            Tcp Print hSocket, "RETR " + remoteFile
            Tcp Line hSocket, buffer
            If buffer_status(buffer) <> 150 Then
               Function = %False
            Else
               '- The total bytes to retrieve is inside
               '  of parenthesis in the returned buffer.
               '
               iTotal = Int(Val(Parse$(buffer, Any "()", 2)))
               iProgress = 0
               '- Read in blocks until we have the whole file.
               Do
                  If read_data_block(hData, buffer) < 1 Then Exit Do
                  Put #iFF,, buffer
                  iProgress = iProgress + Len(buffer)
                  zText = Format$(iProgress) + " of " + Format$(iTotal)
                  '- Progress is passed back via a window-handle
                  If hStatus <> 0 Then
                     SetWindowText hStatus, zText
                  End If
               Loop
               Tcp Line hSocket, buffer
               Function = %True
            End If
            Close hData
         End If
         Close #iFF
      End If
   End If
   set_last_status buffer
End Function
'
'  ftpPutFile
'  Opens a binary mode connection and uploads the specified file to the
'  current working directory on the server.
'  It returns %True on success or %False on failure.
'  hStatus is assumed to be a status window handle if non-zero. As the status
'  of the upload changes, the caption of this window is changed to be the
'  percent complete of the transfer
'==============================================================================
Function ftpPutFile Alias "ftpPutFile" (ByVal hSocket As Long, sServer As String, _
                     ByVal hStatus As Long, remoteFile As String, _
                     localFile As String) Export As Long
   Dim iLoop As Long
   Dim iFF As Long
   Dim iTotal As Long
   Dim iBlocks As Long
   Dim iLeft As Long
   Dim iAbort As Long
   Dim hData As Long
   Dim buffer As String
   Dim zText As Asciiz * 100
   '- If the local file doesn't exist, there's
   '  no point in continuing.
   '
   If Dir$(localFile) = "" Then
      buffer = "001 Local File Not Found"
      Function = %False
   Else
      '- Open the local file and check for error
      iFF = FreeFile
      Open localFile For Binary As #iFF
      If Err Then
         Buffer = Format$(Err, "000") + " Local File Access Error"
         Function = %False
      Else
         '- Open a data socket
         ftpSetBinary hSocket
         hData = get_passive_handle(hSocket, sServer)
         If hData = %INVALID_SOCKET Then
            buffer = Format$(gLastErr, "000") + " " + gLastMsg
            Function = %False
         Else
            '- Request the upload
            Tcp Print hSocket, "STOR " + remoteFile
            Tcp Line hSocket, buffer
            If buffer_status(buffer) <> 150 Then
               Close hData
               Function = %False
            Else
               '- Send the file in chunks
               iTotal = Lof(iFF)
               iBlocks = iTotal \ %FTP_BLOCK_SIZE
               iLeft = iTotal Mod %FTP_BLOCK_SIZE
               iAbort = %False
               For iLoop = 1 To iBlocks
                  buffer = Space$(%FTP_BLOCK_SIZE)
                  Get #iFF,, buffer
                  If Err Then
                     iAbort = %True
                     Exit For
                  End If
                  If write_data_block(hData, buffer) = %False Then
                     iAbort = %True
                     Exit For
                  End If
                  zText = Format$(iLoop * %FTP_BLOCK_SIZE) + " of " + Format$(iTotal)
                  If hStatus > 0 Then
                     SetWindowText hStatus, zText
                  End If
                  'Tcp Print hSocket, "NOOP"
                  'Tcp Line hSocket, buffer
               Next i
               '- Send the last piece
               If (iAbort = %False) And (iLeft > 0) Then
                  buffer = Space$(iLeft)
                  Get #iFF,, buffer
                  If Err Then
                     iAbort = %True
                  Else
                     If write_data_block(hData, buffer) = %False Then
                        iAbort = %True
                     End If
                  End If
               End If
               Close hData
               '- Cleanup and return
               If iAbort Then
                  Function = %False
               Else
                  Tcp Line hSocket, buffer
                  If buffer_status(buffer) = 226 Then
                     Function = %True
                  Else
                     Function = %False
                  End If
               End If
            End If
         End If
         Close #iFF
      End If
   End If
   set_last_status buffer
End Function
'
'  ftpGetList
'  Fills sList() with all of the file info returned
'  by a LIST command. Returns the number of lines
'  returned. (same as ubound(sList))
'==============================================================================
Function ftpGetList Alias "ftpGetList" (ByVal hSocket As Long, _
                  sServer As String, sList() As String) Export As Long
   Dim hData As Long
   Dim buffer As String
   ReDim sList(0 To 0) As String
   '- Get a passive mode socket handle
   hData = get_passive_handle(hSocket, sServer)
   If hData = %INVALID_SOCKET Then
      buffer = Format$(gLastErr, "000") + " " + gLastMsg
      Function = 0
   Else
      '- Request the directory list
      On Error Resume Next
      ftpSetAscii hSocket
      Tcp Print hSocket, "LIST"
      Tcp Line hSocket, buffer
      Do
         Tcp Line hData, buffer
         If Trim$(buffer) = "" Then
            Exit Do
         ElseIf Err Then
            Exit Do
         End If
         ReDim Preserve sList(0 To UBound(sList) + 1) As String
         sList(UBound(sList)) = buffer
      Loop
      Tcp Line hSocket, buffer
      Close hData
      Function = UBound(sList)
   End If
   set_last_status buffer
End Function
'
'  ftpGetFileList
'  Fills the array - sFiles() with a list of all the files in
'  the current directory. Directories are not included in the
'  list. The list is dimmed from 0 to the number of files in
'  the directory. The 0 element is ignored. The function returns
'  ubound(sFiles)
'==============================================================================
Function ftpGetFileList Alias "ftpGetFileList" (ByVal hSocket As Long, _
                  sServer As String, sFiles() As String) Export As Long
   Dim i As Long
   Dim iList As Long
   Dim sList() As String
   ReDim sList(0 To 0) As String
   ReDim sFiles(0 To 0) As String
   iList = ftpGetList(hSocket, sServer, sList())
   If iList > 0 Then
      For i = 1 To iList
         If Left$(sList(i), 1) = "-" Then
            ReDim Preserve sFiles(0 To UBound(sFiles) + 1) As String
            sFiles(UBound(sFiles)) = sList(i)
         End If
      Next i
   End If
   Function = UBound(sFiles)
End Function
'
'  ftpGetDirList
'  Fills the sDirs() array with a list of all the directories
'  in the current directory on the ftp server. It returns the
'  number of elements in sDirs. The array is dimensioned from
'  0 through the number of directories found. The 0 element
'  is left empty and should be ignored.
'==============================================================================
Function ftpGetDirList Alias "ftpGetDirList" _
               (ByVal hSocket As Long, sServer As String, sDirs() As String) Export As Long
   Dim i As Long
   Dim iList As Long
   Dim sList() As String
   ReDim sList(0 To 0) As String
   ReDim sDirs(0 To 0) As String
   iList = ftpGetList(hSocket, sServer, sList())
   If iList > 0 Then
      For i = 1 To iList
         If UCase$(Left$(sList(i), 1)) = "D" Then
            ReDim Preserve sDirs(0 To UBound(sDirs) + 1) As String
            sDirs(UBound(sDirs)) = sList(i)
         End If
      Next i
   End If
   Function = UBound(sDirs)
End Function
'
'  connect to ftp server
'  if successful, a valid socket handle is return
'  if not, %INVALID_SOCKET
'==============================================================================
Function ftpConnect Alias "ftpConnect" (sServer As String, _
                  sUser As String, sPW As String) Export As Long
   Dim hSocket As Long
   Dim buffer As String
   '- Connect to port
   On Error Resume Next
   hSocket = FreeFile
   Tcp Open Port %IPPORT_FTP At sServer As hSocket
   If Err Then
      set_last_status Format$(Err, "000") + " Unable to open socket"
      Function = %INVALID_SOCKET
      '============
      Exit Function
      '============
   End If
   '- What does the server have to say
   Do
      Tcp Line hSocket, buffer
      If buffer_status(buffer) <> 220 Then
         set_last_status buffer
         Close hSocket
         Function = %INVALID_SOCKET
         '============
         Exit Function
         '============
      End If
      If Mid$(buffer, 4, 1) <> "-" Then Exit Do
   Loop
   '- Login with user and password
   Tcp Print hSocket, "USER " + sUser
   Tcp Line hSocket, buffer
   '- Send the password if requested.
   If buffer_status(buffer) = 331 Then
      Tcp Print hSocket, "PASS " + sPW
   End If
   Do
      Tcp Line hSocket, buffer
      If Mid$(buffer, 4, 1) <> "-" Then Exit Do
   Loop
   set_last_status buffer
   If buffer_status(buffer) <> 230 Then
      Close hSocket
      Function = %INVALID_SOCKET
      '============
      Exit Function
      '============
   End If
   Function = hSocket
End Function
'
'  ftpQuit
'  Closes an FTP connection
'==============================================================================
Sub ftpQuit Alias "ftpQuit" (ByVal hSocket As Long) Export
   Dim buffer As String
   Tcp Print hSocket, "QUIT"
   Tcp Line hSocket, buffer
   Do Until Mid$(buffer, 4, 1) <> "-"
      Tcp Line hSocket, buffer
   Loop
   set_last_status buffer
   Close hSocket
End Sub
'
'  ftpTranslateErrorCode
'  Translates an error code returned from an FTP server into
'  a common text interpretation ofthe error. It's usually better
'  to just read the string returned by the server instead of using
'  this function.
'=============================================================================
Function ftpTranslateErrorCode(ByVal errorCode As Long) As String
   Dim e As String
   Select Case errorCode
      Case %FTP_RESTART_MARKER_REPLY
         e = "Restart Marker Reply"
      Case %FTP_SERVICE_READY_SOON
         e = "Service available soon:
      Case %FTP_OPEN_STARTING_TRANSFER
         e = "Data Connection Open. Starting Transfer"
      Case %FTP_FILE_READY
         e = "File ready. About to open data connection."
      Case %FTP_OK
         e = "Success"
      Case %FTP_COMMAND_NOT_IMPLEMENTED
         e = "Command Not Implemented"
      Case %FTP_SYSTEM_STATUS
         e = "System Status"
      Case %FTP_DIRECTORY_STATUS
         e = "Directory Status"
      Case %FTP_FILE_STATUS
         e = "File Status"
      Case %FTP_HELP_MESSAGE
         e = "Help Message"
      Case %FTP_SYSTEM_TYPE_NAME
         e = "System Type Name"
      Case %FTP_SERVICE_READY
         e = "Service Ready"
      Case %FTP_CONNECTION_CLOSING
         e = "Connection Closing. Operation Aborted"
      Case %FTP_OPEN_NO_CURRENT_TRANSFER
         e = "Connection Open. No transfer in progress."
      Case %FTP_TRANSFER_COMPLETE
         e = "Transfer completed successfully."
      Case %FTP_ENTERING_PASSIVE_MODE
         e = "Entering Passive Mode"
      Case %FTP_USER_LOGGED_IN
         e = "User Logged In"
      Case %FTP_ACTION_COMPLETED
         e = "Action Completed Successfully"
      Case %FTP_DIRECTORY_OK
         e = "Directory Created Ok or Current Directory is Ok."
      Case %FTP_PASSWORD_NEEDED
         e = "User Ok, but password needed."
      Case %FTP_ACCOUNT_NEEDED
         e = "Account needed"
      Case %FTP_ACTION_PENDING
         e = "Action Pending Further Information"
      Case %FTP_SERVICE_UNAVAILABLE
         e = "Service Unavailable"
      Case %FTP_DATA_CONN_NOT_OPENED
         e = "Data Connection Not Opened."
      Case %FTP_DATA_CONN_CLOSED
         e = "Data Connection Closed"
      Case %FTP_FILE_UNAVAILABLE
         e = "File unavailable"
      Case %FTP_LOCAL_PROCESSING_ERROR
         e = "Local Processing Error"
      Case %FTP_NOT_ENOUGH_DISK_SPACE
         e = "Not Enough Disk Space"
      Case %FTP_COMMAND_NOT_RECOGNIZED
         e = "Command Not Recognized"
      Case %FTP_BAD_PARAMETERS
         e = "Syntax Error: Invalid parameters or arguments"
      Case %FTP_NOT_IMPLEMENETED
         e = "Not Implemented"
      Case %FTP_BAD_COMMAND_SEQUENCE
         e = "Bad Command Sequence"
      Case %FTP_COMMAND_UNAVAILABLE
         e = "Command Unavailable"
      Case %FTP_USER_NOT_LOGGED_IN
         e = "User Not Logged In"
      Case %FTP_ACCOUNT_NEEDED_TO_STORE
         e = "Account Needed to Store Files"
      Case %FTP_ACCESS_DENIED
         e = "Access Denied"
      Case %FTP_PAGE_TYPE_UNKNOWN
         e = "Page Type Unknown"
      Case %FTP_STORAGE_LIMIT_EXCEEDED
         e = "Storage Limit Exceeded"
      Case %FTP_BAD_FILE_NAME
         e = "Bad File Name"
   End Select
   Function = e
End Function
